/* tcps.c - tcp server
 * Time-stamp: <2004-09-06 njk>
 *  
 * (C) 2004 Nicholas J. Kain <njk@aerifal.cx>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>

#include <pwd.h>
#include <grp.h>

#include <signal.h>
#include <errno.h>

#define _GNU_SOURCE
#include <getopt.h>

#include "defines.h"
#include "log.h"
#include "chroot.h"
#include "pidfile.h"
#include "signals.h"
#include "nstrl.h"
#include "cap.h"
#include "exec.h"

int numchildren = 0;
static int backlog = 30;
static int maxconcurrent = 10;
static int port;
static char *cmd = NULL, *args = NULL;

static struct in_addr addy;

static volatile sig_atomic_t pending_exit, pending_reap;

static void sighandler(int sig) {
	switch (sig) {
		case SIGTERM:
		case SIGINT:
		    pending_exit = 1;
			break;
		case SIGCHLD:
			pending_reap = 1;
			break;
	}
}

static void fix_signals(void) {
  disable_signal(SIGPIPE);
  disable_signal(SIGUSR1);
  disable_signal(SIGUSR2);
  disable_signal(SIGTSTP);
  disable_signal(SIGTTIN);
  disable_signal(SIGHUP);
  
  hook_signal(SIGCHLD, sighandler, 0);
  hook_signal(SIGINT, sighandler, 0);
  hook_signal(SIGTERM, sighandler, 0);
}

static void suicide(char *errmsg, const char *perrmsg, int status)
{
    if (errmsg) log_line(errmsg);
    if (!gflags_detach && perrmsg) perror(perrmsg);
    exit(status);
}

static void grim_reaper(void)
{
	while (waitpid(-1, NULL, WNOHANG) > 0)
		numchildren--;
	pending_reap = 0;
}

static void handle_signals(void)
{
	if (pending_reap)
		grim_reaper();
	if (pending_exit)
		exit(EXIT_SUCCESS);
}

/* Opens a non-blocking listening socket with the appropriate properties. */
static int get_listen()
{
	int lsock, ret;
	struct sockaddr_in sa;

	if (port == 0)
		suicide("FATAL - no listen port specified\n", "get_listen",
				EXIT_FAILURE);

	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	sa.sin_addr = addy;
	
	lsock = socket(PF_INET, SOCK_STREAM, 0);
	if (lsock == -1)
		suicide("FATAL - failed to create socket\n",
				"get_listen - socket", EXIT_FAILURE);

	fcntl(lsock, F_SETFL, O_NONBLOCK);

	ret = bind(lsock, (struct sockaddr *) &sa, sizeof sa);
	if (ret)
		suicide("FATAL - failed to bind socket\n",
				"get_listen - bind", EXIT_FAILURE);
	ret = listen(lsock, backlog);
	if (ret)
		suicide("FATAL - failed to listen on socket\n",
				"get_listen - listen", EXIT_FAILURE);

	return lsock;
}

/* Abstracts away the details of accept()ing a socket connection. */
static void accept_conns(int lsock)
{
	int fd;
	struct sockaddr_in sock_addr;
	socklen_t sock_len = sizeof sock_addr;

	for(;;)
	{
		fd = accept(lsock, (struct sockaddr *) &sock_addr, &sock_len);

		if (fd != -1) {
			if (numchildren < maxconcurrent) {
				exec_and_fork(fd, cmd, args);
				return;
			} else {
				for(;;) {
					if (numchildren < maxconcurrent) {
						exec_and_fork(fd, cmd, args);
						return;
					}
					sleep(60);
					handle_signals();
				}
				return;
			}
		}

		switch (errno) {
			case EAGAIN:
#ifdef LINUX							
			case ENETDOWN:
			case EPROTO:
			case ENOPROTOOPT:
			case EHOSTDOWN:
			case ENONET:
			case EHOSTUNREACH:
			case EOPNOTSUPP:
			case ENETUNREACH:
#endif							
				return;

			case EINTR:
				continue;

			case EBADF:
			case ENOTSOCK:
			case EINVAL:
			case ECONNABORTED:
			case EMFILE:
			case ENFILE:
				log_line("warning: accept returned %s!\n", strerror(errno));
				return;

			default:
				log_line("warning: accept returned a mysterious error: %s\n",
						strerror(errno));
				return;
		}
	}
}

static void dispatch_work(int lsock) 
{
    fd_set rfds;
   
	for (;;) {
		handle_signals();

		FD_ZERO(&rfds);
        FD_SET(lsock, &rfds);

		if (select(lsock + 1, &rfds, NULL, NULL, NULL) == -1) {
			if (errno == EINTR)
				continue;
			if (pending_exit == 1)
				return;
			suicide("FATAL - select returned an error!\n",
					"dispatch_work - select", EXIT_FAILURE);
		}

        /* handle pending connections */
		if (FD_ISSET(lsock, &rfds))
			accept_conns(lsock);
    }
}

int main(int argc, char** argv) {
	int c, t, uid = 0, gid = 0, i, len, socket;
	char *pidfile = NULL;
	char *chrootd = NULL;
	char *capability = NULL;
	char *p;
	struct passwd *pws;
	struct group *grp;

	inet_aton("0.0.0.0", &addy);
	
	while (1) {
		int option_index = 0;
		static struct option long_options[] = {
			{"detach", 0, 0, 'd'},
			{"nodetach", 0, 0, 'n'},
			{"pidfile", 1, 0, 'f'},
			{"quiet", 0, 0, 'q'},
			{"chroot", 1, 0, 'c'},
			{"capability", 1, 0, 'C'},
			{"servermax", 1, 0, 's'},
			{"backlog", 1, 0, 'b'},
			{"address", 1, 0, 'a'},
			{"port", 1, 0, 'p'},
			{"user", 1, 0, 'u'},
			{"group", 1, 0, 'g'},
			{"help", 0, 0, 'h'},
			{"version", 0, 0, 'v'},
			{0, 0, 0, 0}
		};

		c = getopt_long (argc, argv, "dnf:qc:C:s:b:a:p:ou:g:hv", long_options, &option_index);
		if (c == -1) break;

		switch (c) {

			case 'h':
printf("tcps %s, tcp server.  Licensed under GNU GPL.\n", TCPS_VERSION);
				printf(
"Copyright (C) 2004 Nicholas J. Kain\n"
"Usage: tcps [OPTIONS]\n"
"  -d, --detach                detach from TTY and daemonize (default)\n"
"  -n, --nodetach              stay attached to TTY\n"
"  -q, --quiet                 don't print to std(out|err) or log\n"
"  -c, --chroot                path where ifchd should chroot\n"
"  -C, --capability            capability string to use when dropping root\n");
printf(
"  -s, --servermax             maximum number of servers to allow at once\n"
"  -b, --backlog               maximum simultaneous connections accepted\n"
"  -a, --address               address on which to listen (default 0.0.0.0)\n"
"  -p, --port                  port on which to listen\n"
"  -f, --pidfile               pidfile path\n");
printf(            
"  -u, --user                  user name that tcps should run as\n"
"  -g, --group                 group name that tcps should run as\n"
"  -h, --help                  print this help and exit\n"
"  -v, --version               print version information and exit\n");
				exit(EXIT_FAILURE);
				break;

			case 'v':
				printf("tcps %s, tcp server.  Licensed under GNU GPL.\n", TCPS_VERSION);
				printf(
"Copyright (C) 2004 Nicholas J. Kain\n"
"This is free software; see the source for copying conditions.  There is NO\n"
"WARRANTY; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
				exit(EXIT_FAILURE);
				break;

			case 'd':
				gflags_detach = 1; 
				break;

			case 'n':
				gflags_detach = 0;
				break;

			case 'q':
				gflags_quiet = 1;
				break;

			case 's':
				maxconcurrent = atoi(optarg);
				break;

			case 'b':
				backlog = atoi(optarg);
				break;

			case 'p':
				port = atoi(optarg);
				break;

			case 'c':
				len = strlen(optarg) + 1;
				free(chrootd);
				chrootd = malloc(len);
				strlcpy(chrootd, optarg, len);
				break;

			case 'C':
				len = strlen(optarg) + 1;
				free(capability);
				capability = malloc(len);
				strlcpy(capability, optarg, len);
				break;

			case 'f':
				len = strlen(optarg) + 1;
				free(pidfile);
				pidfile = malloc(len);
				strlcpy(pidfile, optarg, len);
				break;


			case 'a':
				inet_aton(optarg, &addy);
				break;
				
			case 'u':
				t = (unsigned int) strtol(optarg, &p, 10);
				if (*p != '\0') {
					pws = getpwnam(optarg);
					if (pws) {
						uid = (int)pws->pw_uid;
						if (!gid)
							gid = (int)pws->pw_gid;
					} else suicide("FATAL - Invalid uid specified.\n", NULL,
							EXIT_FAILURE);
				} else 
					uid = t;
				break;

			case 'g':
				t = (unsigned int) strtol(optarg, &p, 10);
				if (*p != '\0') {
					grp = getgrnam(optarg);
					if (grp) {
						gid = (int)grp->gr_gid;
					} else suicide("FATAL - Invalid gid specified.\n", NULL,
							EXIT_FAILURE);
				} else
					gid = t;
				break;
		}
	}

	if (argv[optind] != NULL) {
		len = strlen(argv[optind]) + 1;
		cmd = malloc(len);
		strlcpy(cmd, argv[optind], len);

		for (len = 0, i = optind + 1; argv[i] != NULL; ++i)
			len += strlen(argv[i]) + 1; /* +1 accounts for space AND '\0' */
		if (len) {
			args = malloc(len);
			i = optind + 1;
			strlcpy(args, argv[i++], len);
			while (argv[i] != NULL) {
				strlcat(args, " ", len);
				strlcat(args, argv[i++], len);
			}
		}
	}
	if (gflags_detach)
		if (daemon(0,0)) {
			log_line("FATAL - detaching fork failed\n");
			exit(EXIT_FAILURE);
		}

	if (!cmd) {
		log_line("FATAL - no server daemon to run!\n");
		exit(EXIT_FAILURE);
	}

	if (!port) {
		log_line("FATAL - no listening port specified!\n");
		exit(EXIT_FAILURE);
	}
	
	if (file_exists(pidfile, "w"))
		write_pid(pidfile);

	umask(077);
	fix_signals();
	fix_env(uid);
	socket = get_listen();
	
	imprison(chrootd);
	drop_root(uid, gid, capability);

	/* Cover our tracks... */
	free(chrootd);
	free(pidfile);
	free(capability);
	chrootd = NULL;
	pidfile = NULL;
	capability = NULL;

	close(0);
	close(1);
	
	dispatch_work(socket);

	exit(EXIT_SUCCESS);
}

